@@ -24,72 +24,122 @@
 # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 # OTHER DEALINGS IN THE SOFTWARE.
 import re
+import os
+import sys
 import inspect
 import traceback
 from datetime import datetime
+from functools import wraps
 from pprint import pformat
-from threading import local
 from copy import deepcopy
-version = '0.6.1'
+from collections import Iterable
+version = '0.7.0'
 
 
 def itemize_length(items):
     length = len(items)
     return '%d item%s' % (length, length > 1 and "s" or "")
 
+not_here_error = \
+    'you have tried to access the attribute "%s" the context ' \
+    '(aka VariablesBag), but there is no such attribute assigned to it. ' \
+    'Maybe you misspelled it ? Well, here are the options: [%s]'
+
+
+class VariablesBag(dict):
+    __varnames__ = None
+    __sure_actions_ran__ = None
+    __sure_action_results__ = None
+
+    def __init__(self, *args, **kw):
+        self.__varnames__ = []
+        self.__sure_actions_ran__ = []
+        self.__sure_action_results__ = []
+        return super(VariablesBag, self).__init__(*args, **kw)
+
+    def __setattr__(self, attr, value):
+        if attr not in dir(VariablesBag):
+            self[attr] = value
+            self.__varnames__.append(attr)
+        return super(VariablesBag, self).__setattr__(attr, value)
+
+    def __getattr__(self, attr):
+        try:
+            return super(VariablesBag, self).__getattribute__(attr)
+        except AttributeError:
+            if attr not in dir(VariablesBag):
+                raise AssertionError(not_here_error % (
+                    attr,
+                    ", ".join(map(repr, self.__varnames__)),
+                ))
+
 
 class CallBack(object):
+    context_error = "the function %s defined at %s line %d, is being "\
+        "decorated by either @that_with_context or @scenario, so it should " \
+        "take at least 1 parameter, which is the test context"
+
     def __init__(self, cb, args, kwargs):
         self.callback = cb
         self.args = args or []
         self.kwargs = kwargs or {}
+        self.callback_name = cb.__name__
+        self.callback_filename = os.path.split(cb.func_code.co_filename)[-1]
+        self.callback_lineno = cb.func_code.co_firstlineno + 1
 
     def apply(self, *optional_args):
-        if not callable(self.callback):
-            return
-
         args = list(optional_args)
         args.extend(self.args)
         try:
-            self.callback(*args, **self.kwargs)
-        except TypeError:
-            neat_args = [a for a in args if a not in optional_args]
-            self.callback(*neat_args, **self.kwargs)
+            return self.callback(*args, **self.kwargs)
+        except:
+            err = unicode(sys.exc_info()[1])
+            if err.startswith(self.callback_name) and \
+               'takes no arguments (1 given)' in err:
+                raise TypeError(self.context_error % (
+                    self.callback_name,
+                    self.callback_filename,
+                    self.callback_lineno,
+                    )
+                )
+            raise
 
 
 def that_with_context(setup=None, teardown=None):
     def dec(func):
+        @wraps(func)
         def wrap(*args, **kw):
-            context = local()
+            context = VariablesBag()
 
             if callable(setup):
                 cb = CallBack(setup, args, kw)
                 cb.apply(context)
 
-            elif isinstance(setup, (set, list, tuple)):
+            elif isinstance(setup, Iterable):
                 for s in setup:
                     cb = CallBack(s, args, kw)
                     cb.apply(context)
 
             test = CallBack(func, args, kw)
             try:
-                test.apply(context)
+                res = test.apply(context)
             finally:
                 if callable(teardown):
                     cb = CallBack(teardown, args, kw)
                     cb.apply(context)
 
-                elif isinstance(teardown, (set, list, tuple)):
+                elif isinstance(teardown, Iterable):
                     for s in teardown:
                         cb = CallBack(s, args, kw)
                         cb.apply(context)
 
-        wrap.__name__ = func.__name__
-        wrap.__doc__ = func.__doc__
+            return res
         return wrap
 
     return dec
 
+scenario = that_with_context
+
 
 def explanation(msg):
     def dec(func):
@@ -175,7 +225,9 @@ class that(object):
 
             if isinstance(exc, type) and issubclass(exc, Exception):
                 if not isinstance(e, exc):
-                    raise AssertionError('%r should raise %r, but raised %r' % (self._src, exc, e.__class__))
+                    raise AssertionError(
+                        '%r should raise %r, but raised %r' % (
+                            self._src, exc, e.__class__))
 
                 if isinstance(msg, basestring) and msg not in unicode(e):
                     raise AssertionError('%r raised %s, but the exception message does not match. Expected %r, got %r' % (self._src, e, msg, unicode(e)))
@@ -307,7 +359,7 @@ class that(object):
             error = 'the length of %r should be greater then or equals %d, but is %d' % (
                 self._src,
                 that,
-                length
+                length,
             )
             raise AssertionError(error)
 
@@ -321,7 +373,7 @@ class that(object):
             error = 'the length of %r should be lower then %d, but is %d' % (
                 self._src,
                 that,
-                length
+                length,
             )
             raise AssertionError(error)
 
@@ -331,14 +383,15 @@ class that(object):
         that = self._get_that(that)
 
         length = len(self._src)
+        error = 'the length of %r should be lower then or equals %d, but is %d'
 
         if length > that:
-            error = 'the length of %r should be lower then or equals %d, but is %d' % (
+            msg = error % (
                 self._src,
                 that,
-                length
+                length,
             )
-            raise AssertionError(error)
+            raise AssertionError(msg)
 
         return True
 
@@ -350,7 +403,7 @@ class that(object):
             error = 'the length of %r should be %d, but is %d' % (
                 self._src,
                 that,
-                length
+                length,
             )
             raise AssertionError(error)
 
@@ -370,7 +423,7 @@ class that(object):
     def matches(self, items):
         msg = '%r[%d].%s should be %r, but is %r'
         get_eval = lambda item: eval(
-            "%s.%s" % ('current', self._eval), {}, {'current': item}
+            "%s.%s" % ('current', self._eval), {}, {'current': item},
         )
 
         if self._eval and is_iterable(self._src):
@@ -384,7 +437,7 @@ class that(object):
 
                     raise AssertionError(
                         '%r has %d items, but the matching list has %d: %r'
-                        % (source, source_len, items_len, items)
+                        % (source, source_len, items_len, items),
                     )
 
             for index, (item, other) in enumerate(zip(self._src, items)):
@@ -407,7 +460,9 @@ class that(object):
         try:
             lst = list(self._src)
             length = len(lst)
-            assert length == 0, '%r is not empty, it has %s' % (self._src, itemize_length(self._src))
+            assert length == 0, \
+                   '%r is not empty, it has %s' % (self._src,
+                                                   itemize_length(self._src))
             return True
 
         except TypeError:
@@ -430,11 +485,18 @@ class that(object):
         return what in items
 
     def contains(self, what):
-        assert isinstance(what, basestring), '%r should be a string' % what
-        assert isinstance(self._src, basestring), '%r is not a string, so is is impossible to check if "%s" is there' % (self._src, what)
-        assert what in self._src, '"%s" should be in "%s"' % (what, self._src)
+        assert what in self._src, '%r should be in %r' % (what, self._src)
         return True
 
+    def does_not_contain(self, what):
+        assert what not in self._src, \
+            '%r should NOT be in %r' % (what, self._src)
+
+        return True
+
+    doesnt_contain = does_not_contain
+
+
 def within(**units):
     assert len(units) == 1, 'use within(number=unit). e.g.: within(one=second)'
 
@@ -444,6 +506,7 @@ def within(**units):
     convert_from, convert_to = UNITS[unit]
     timeout = convert_from(value)
     exc = []
+
     def dec(func):
         def wrap(*args, **kw):
             start = datetime.utcnow()
@@ -478,10 +541,22 @@ def within(**units):
     return dec
 
 UNITS = {
-    'minutes': (lambda from_num: from_num / 60.0, lambda to_num: to_num * 6000000),
-    'seconds': (lambda from_num: from_num, lambda to_num: to_num / 100000),
-    'miliseconds': (lambda from_num:from_num * 1000, lambda to_num: to_num/100),
-    'microseconds': (lambda from_num:from_num * 100000, lambda to_num: to_num),
+    'minutes': (
+        lambda from_num: from_num / 60.0,
+        lambda to_num: to_num * 6000000,
+    ),
+    'seconds': (
+        lambda from_num: from_num,
+        lambda to_num: to_num / 100000,
+    ),
+    'miliseconds': (
+        lambda from_num: from_num * 1000,
+        lambda to_num: to_num / 100,
+    ),
+    'microseconds': (
+        lambda from_num: from_num * 100000,
+        lambda to_num: to_num,
+    ),
 }
 
 milisecond = miliseconds = 'miliseconds'
@@ -489,6 +564,7 @@ microsecond = microseconds = 'microseconds'
 second = seconds = 'seconds'
 minute = minutes = 'minutes'
 
+
 def word_to_number(word):
     basic = {
         'one': 1,
@@ -510,3 +586,30 @@ def word_to_number(word):
         raise AssertionError(
             'sure supports only literal numbers from one to twelve, ' \
             'you tried the word "twenty"')
+
+
+def action_in(scenario):
+    def decorate_and_absorb(func):
+        @wraps(func)
+        def wrapper(*args, **kw):
+            scenario.__sure_actions_ran__.append((func, args, kw))
+            result = func(*args, **kw)
+            scenario.__sure_action_results__.append(result)
+
+            return scenario
+
+        setattr(scenario, func.__name__, wrapper)
+        return wrapper
+
+    if not hasattr(scenario, 'contextualized_as'):
+
+        def contextualized_as(name):
+            last_result = scenario.__sure_action_results__[-1]
+            setattr(scenario, name, last_result)
+            return scenario
+
+        scenario.contextualized_as = contextualized_as
+
+    return decorate_and_absorb
+
+action_for = action_in
